#! /usr/bin/perl

# This script prepares a request for the POLYVIEW webserver,
# submits the request and then processes the response.

# All parameters for the script (request) should be specified in
# the options.txt file located at the working directory.

# The script and options file were written by
# Alexey Porollo
# Department of Environmental Health
# University of Cincinnati College of Medicine
# and
# Division of Biomedical Informatics
# Children's Hospital Research Foundation
#
# Cincinnati, OH
# USA

# USAGE: perl polyview.pl

use LWP::UserAgent;

use constant OPTFILE => 'options.txt';
use constant ADDRESS => 'http://polyview.cchmc.org/cgi-bin/pr_picture.cgi';
use constant FILEADR => 'http://polyview.cchmc.org/cgi-bin/get_data.cgi?FN=';

# read options file
my %params = ReadOptions();
$params{OUTPUTDIRECTORY} = '.' unless $params{OUTPUTDIRECTORY};
my $t_info = ($params{INFORMATION} =~ /image|sequence|both/i) ? lc($&) : 'image';

# fill up the array of hashes with data to be submitted
my @str;
my $id_count = 0; # unnamed structures counter

# take explicitly specified sequence annotation first
if (exists $params{AASEQUENCE} && length($params{AASEQUENCE}) > 0) {
    my $pname = ($params{PROTEINNAME}) ? $params{PROTEINNAME} :
	                                 'Noname' . ++$id_count;
    my %annot = ('Name' => $pname);
    $annot{AA}   = $params{AASEQUENCE};
    $annot{SS}   = $params{SSSEQUENCE} if exists $params{SSSEQUENCE};
    $annot{SSC}  = $params{SSCONFIDENCE} if exists $params{SSCONFIDENCE};
    $annot{SEA}  = $params{SEASEQUENCE} if exists $params{SEASEQUENCE};
    $annot{SEAC} = $params{SEACONFIDENCE} if exists $params{SEACONFIDENCE};
    $annot{ARB}  = $params{ARBSEQUENCE} if exists $params{ARBSEQUENCE};
    push @str, {%annot};
}
# take structure annotations from the input file
if (exists $params{INPUTFILENAME} && open(IN, $params{INPUTFILENAME})) {
    while (<IN>) {
    	# get protein name
    	chomp;
	    $_ =~ s/\r//; # for dos prepared files
    	my $pname = ($_) ? $_ : 'Noname' . ++$id_count;
	    my %annot = ('Name' => $pname);
    	# get amino acid sequence
    	$_ = <IN>;
	    chomp;
    	$_ =~ s/\r//;
	    $annot{AA} = $_;
        # get secondary structure
    	$_ = <IN>;
	    chomp;
    	$_ =~ s/\r//;
	    $annot{SS} = $_ if $_;
        # get secondary structure confidence
    	$_ = <IN>;
	    chomp;
    	$_ =~ s/\r//;
	    $annot{SSC} = $_ if $_;
        # get relative solvent accessibility
    	$_ = <IN>;
	    chomp;
    	$_ =~ s/\r//;
	    $annot{SEA} = $_ if $_;
        # get RSA confidence
	    $_ = <IN>;
    	chomp;
	    $_ =~ s/\r//;
    	$annot{SEAC} = $_ if $_;
        # get arbitrary sequence profile
	    $_ = <IN>;
    	chomp;
	    $_ =~ s/\r//;
    	$annot{ARB} = $_ if $_;
        # skip reserved lines
	    $_ = <IN>;
	    # store annotation
    	push @str, {%annot};
    }
    close IN;
}
# add PDB-based entries in the end
if (exists $params{PDBCODE}) {
    my %annot = ('Name' => $params{PDBCODE});
    $annot{PDB} = $params{PDBCODE};
    push @str, {%annot};
}
if (exists $params{PDBFILENAME}) {
    my %annot = ('Name' => 'Custom file');
    $annot{PDBfile} = $params{PDBFILENAME};
    push @str, {%annot};
}

# define agent
my $ua = LWP::UserAgent->new;
$ua->agent("POLYVIEW off-line user");

# submit proteins
$count = $suc_count = 0;
for my $i (0 .. $#str) {
    $count++;
    print "Submitting query: $str[$i]{Name}...\n";

    # check if protein is valid
    if (!exists $str[$i]{PDB} &&
	    (exists $str[$i]{SS} && length($str[$i]{SS}) != length($str[$i]{AA})) ||
	    (exists $str[$i]{SSC} && length($str[$i]{SSC}) != length($str[$i]{AA})) ||
	    (exists $str[$i]{SEA} && length($str[$i]{SEA}) != length($str[$i]{AA})) ||
	    (exists $str[$i]{SEAC} && length($str[$i]{SEAC}) != length($str[$i]{AA})) ||
	    (exists $str[$i]{ARB} && length($str[$i]{ARB}) != length($str[$i]{AA}))) {
	    print "Lengths of sequences are inconsistent for protein $str[$i]{Name}! Skipped...\n";
	    next;
    }

    # prepare a request string
    my %req_par = ();
    if (exists $str[$i]{PDB}) {
	    $req_par{PDBName} = $str[$i]{PDB};
    }
    elsif (exists $str[$i]{PDBfile} && -f $str[$i]{PDBfile}) {
	    $req_par{PDBFileName} = ["$str[$i]{PDBfile}"];
    }
    elsif (length($str[$i]{AA}) > 0) {
	    $req_par{seqName}    = $str[$i]{Name};
	    $req_par{aaSeq}      = $str[$i]{AA};
	    $req_par{ssSeq}      = $str[$i]{SS} if exists $str[$i]{SS};
	    $req_par{confSeq}    = $str[$i]{SSC} if exists $str[$i]{SSC};
	    $req_par{seaSeq}     = $str[$i]{SEA} if exists $str[$i]{SEA};
	    $req_par{seaConfSeq} = $str[$i]{SEAC} if exists $str[$i]{SEAC};
	    $req_par{arbitSeq}   = $str[$i]{ARB} if exists $str[$i]{ARB};
    }
    else {
	    print "No amino acid sequence specified for $str[$i]{Name}! Skipped...\n";
	    next;
    }

    for my $par (sort keys %params) {
	    next if $par =~ /PDBCODE|PROTEINNAME|(AA|SS|SEA|ARB)SEQUENCE|(SS|SEA)CONFIDENCE|OUTPUTDIRECTORY/;

	    $req_par{Model}  = $params{$par} if $par eq 'PRESENTATIONMODEL';
	    $req_par{ColorR} = $params{$par} if $par eq 'BGCRED';
	    $req_par{ColorG} = $params{$par} if $par eq 'BGCGREEN';
	    $req_par{ColorB} = $params{$par} if $par eq 'BGCBLUE';

	    $req_par{VSHideNumbers}        = 1 if $par eq 'HIDERESIDUENUMBERING' && $params{$par};
	    $req_par{VSHideAA}             = 1 if $par eq 'HIDEAASEQUENCE' && $params{$par};
	    $req_par{VSHidePictSS}         = 1 if $par eq 'HIDEGRAPHSS' && $params{$par};
	    $req_par{VSHidePictConfidence} = 1 if $par eq 'HIDEGRAPHCONFSS' && $params{$par};
	    $req_par{VSHideSEA}            = 1 if $par eq 'HIDERSA' && $params{$par};
    	$req_par{VSShowProfile}        = 1 if $par eq 'SHOWCHEMPROFILE' && $params{$par};
	    $req_par{VSShowSS}             = 1 if $par eq 'SHOWLETSS' && $params{$par};
	    $req_par{VSShowConfidence}     = 1 if $par eq 'SHOWSSCONFIDENCE' && $params{$par};
	    $req_par{VSShowSEANum}         = 1 if $par eq 'SHOWNUMRSA' && $params{$par};

	    $req_par{NRL}          = $params{$par} if $par eq 'NUMRESIDUESPERLINE';
	    $req_par{SNR}          = $params{$par} if $par eq 'STARTNUM';
	    $req_par{HRN}          = $params{$par} if $par eq 'HIGHLIGHTRESIDUES';
	    $req_par{HRN_TM}       = $params{$par} if $par eq 'HIGHLIGHTTMRESIDUES';
	    $req_par{HRSSBridge}   = 1 if $par eq 'HLSSBRIDGERESIDUES' && $params{$par};
	    $req_par{HRInt}        = 1 if $par eq 'HLINTERFACIALRESIDUES' && $params{$par};
	    $req_par{DSSP_rsamark} = 1 if $par eq 'HLRSAPROBLEMATICRESIDUES' && $params{$par};
    }
    # submit a request
    my $res = $ua->post(ADDRESS, Content_Type => 'form-data', Content => \%req_par);

    my $resp;
    my @response = ();
    # check the outcome of the response
    if ($res->is_success) {
        $resp = $res->content;
        @response = split "\n", $resp;
        $suc_count++;
    }
    else {
        print "Error: ", $res->status_line, "\n";
        next;
    }

    # get the images
    if ($t_info =~ /image|both/) {
        my $id = 0;
        my $qname = '';
        my $qsfx = '';
        for my $r (@response) {
            # get the protein name
	        if ($r =~ /"Query name: (.*?)"/i) {
	            $qname = $1;
	            $qname =~ s/\s+$//;
	            $qname =~ s/([^\w.-])/_/g;
	        }
            # get link to its picture
    	    if ($r =~ /<img (.*?)src="(.*?)(">)/i) {
    	        if ($2 =~ /FN=(\d+)/) {
    	    	    my $fname = $1; # name of a file with the image
        	    	print "Getting file: $fname... ";
    	        	if (!$qname) {
    	        	    $qname = 'Noname';
    	        	    $qsfx = '_' . ++$id;
    	        	}

                    # make request to get a file with the image
    	        	$req = HTTP::Request->new(GET => FILEADR . $fname);
    	        	my $res_file = $ua->request($req);
    	        	# check the outcome of the response
    	        	if ($res_file->is_success) {
                        # save file on local disk
    	        	    my $out_fname = "$qname$qsfx.gif";
    	        	    while (-f "$params{OUTPUTDIRECTORY}/$out_fname") {
    	        		    $qsfx = '_' . ++$id;
    	        		    $out_fname = "$qname$qsfx.gif";
    	        	    }
    	        	    open(GIF, ">$out_fname") || die "Can't create local file $out_fname: $!\n";
    	        	    binmode GIF;
    	        	    print GIF $res_file->content;
    	        	    close GIF;
    	        	    print "Saved to: $params{OUTPUTDIRECTORY}/$out_fname\n";
    	        	} 
    	        	else {
    	        	    print "Error: ", $res_file->status_line, "\n";
    	        	    next;
    	        	}

    	    	    $qname = '';
    	        }
	        }
        } # for my $r (@response)
    } # if ($t_info =~ /image|both/)
    # get the sequences
    if ($t_info =~ /sequence|both/ && $resp) {
        print "Parsing text for protein sequences...";
	    my %seq;
	    while ($resp =~ /<form name="SequenceForm(\d+)">(.*?)<\/form>/gms) {
	        my $seq_id = $1;
	        my $seq_info = $2;
            my $qn = ($seq_info =~ /Query name: (.*?)">/ms) ? $1 : 'Noname';
            my $aa = ($seq_info =~ /aaSeq" value="(.*?)">/) ? $1 : '';
            my $ss = ($seq_info =~ /ssSeq" value="(.*?)">/) ? $1 : '';
            my $rsa = ($seq_info =~ /seaSeq" value="(.*?)">/) ? $1 : '';
            %{$seq{$seq_id}} = ('Name'=>$qn, 'AA'=>$aa, 'SS'=>$ss, 'RSA'=>$rsa);
            print " $qn";
	    }
	    print "\n";
	    my $out_fname = 'query' . ($i+1) . '.fasta';
	    open(FASTA, ">$out_fname") || die "Can't create a local file $out_fname: $!\n";
	    for my $id (sort {$a<=>$b} keys %seq) {
	        print FASTA ">$seq{$id}{Name}\n$seq{$id}{AA}\n$seq{$id}{SS}\n$seq{$id}{RSA}\n";
	    }
	    close FASTA;
    } # if ($t_info =~ /sequence|both/)

    sleep 5; # to reduce traffic to the server
} # for my $i (0 .. $#str)

print "Total proteins considered=$count successfully submitted=$suc_count\n";
print "Retrieved sequences are in file(s) query*.fasta\n"
    if ($t_info =~ /sequence|both/);

# ReadOptions reads file options.txt
# and creates hash of parameters.
# no INPUT
# RETURN: hash parameter-value
sub ReadOptions {
    my %res = ();

    open(OPT, OPTFILE) || return %res;
    while (<OPT>) {
	    chomp;
	    next if $_ =~ /^\#/;
	    $_ =~ s/ //g;

	    my @data = split '=';
	    next if scalar(@data) <= 1;
        $data[0] = "\U$data[0]";
        $data[1] = 0 if ($data[1] =~ /No|False/i);
        $data[1] = 1 if ($data[1] =~ /Yes|True/i);
	    $res{$data[0]} = $data[1];
    }

    return %res;
}

